/********************************************************                                          
                                            2009 1/1 
                                         2011 3/19 update
  Linux KeyHoleTV (C) 2008,2009           OISEYER Inc.
                                        Takashi Kosaka

   Linux ALSA Sound Dynamic Link Library

   File:                 AlsaObj.c
   Dynamic Link Library: AlsaObj.so

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:
   
   - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
   
   - Neither the name of the Xiph.org Foundation nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.
   
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
************************************************************/
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <unistd.h>
#include <sys/select.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <alsa/asoundlib.h>

#define   MAX_AUDIO_NUM                 20
#define   AUDIO_44100_100_NUM         10240    /* 100 mill sec 44100 Hz Bytes */

/*****************************************************************
  Sound Object
  Interface structure (Do not change)
*****************************************************************/
typedef struct _SoundObj {
  void        *Sample;	/* used by Linux KeyHoleTV */
  void        *Handle;

  unsigned char *(*GetAudioBuffer)(void *OO);
  void           (*SetNextWriteBuffer)(void *OO,int Size);
  void           (*SoundFunction)(void *OO);
  void           (*StopAuidoFunction)(void *OO);
  void           (*FinishAuidoFunction)(void *OO);

} DataSoundObj, *PSoundObj;

/*****************************************************************
 Audio Buffer form KeyHoleTV
*****************************************************************/
typedef struct _AudioBuffer {
  unsigned char *Buf;   //[AUDIO_NOTIFY_SIZE];
  int            Size;
  int            Written;
  struct _AudioBuffer *Next;
} AudioBuffer, *PAudioBuffer;

/*********************************************************
  ALSA $B$rA`:n$9$k%*%V%8%'%/%H(B
*********************************************************/
typedef struct _AlSaObj {
  char               DeviceName[1024];

  snd_pcm_t           *handle; /* AlSA */
  snd_pcm_sw_params_t *sw_params;

  PAudioBuffer Read;
  PAudioBuffer Write;

  unsigned int        Start;
	
  unsigned int        Store;
  unsigned int        Used;
  char      *buf; 
  int        Size;
  int        MaxSize;

  unsigned int        StopIDW;
  unsigned int        StopIDR;

  snd_pcm_sframes_t buffer_size;
  snd_pcm_sframes_t period_size;

  char     *Exiting;
  char     *JoinRoom;

} AlsaObj, *PAlsaObj;

/**********************************************************
  Set HW Param and Softparam
**********************************************************/
static int SetALSADrive(PAlsaObj Obj)
{
  unsigned int         rate;
  snd_pcm_hw_params_t *hw_params;
  unsigned int buffer_time = 1000000;		/* ring buffer length in us */
  unsigned int period_time =  100000;		/* period time in us        */
  int                  err,dir;
	
  rate = 44100;

  /* $B:G=i$K(B ALSA $B$r3+$$$F$_$k(B */
  if ((err = snd_pcm_open (&Obj->handle,Obj->DeviceName,
			   SND_PCM_STREAM_PLAYBACK , 0)) < 0) {
    //fprintf(stderr,"snd_pcm_open\n");
    return(-1);
  }


  if ((err = snd_pcm_hw_params_malloc (&hw_params)) < 0) {
    //fprintf(stderr,"snd_pcm_hw_params_malloc\n");
    return(-1);
  }

  if ((err = snd_pcm_hw_params_any (Obj->handle, hw_params)) < 0) {
    //fprintf(stderr,"snd_pcm_hw_params_any\n");
    return(-1);
  }

  if ((err = snd_pcm_hw_params_set_access(Obj->handle, 
					  hw_params, 
					  SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
    //fprintf(stderr,"snd_pcm_hw_params_set_access\n");
    return(-1);
  }

  if((err = snd_pcm_hw_params_set_rate_resample(Obj->handle,hw_params ,1)) < 0 ) {
    return(-1);
  }

  if ((err = snd_pcm_hw_params_set_format (Obj->handle, hw_params, 
					   SND_PCM_FORMAT_S16_LE)) < 0) {
    return(-1);
  }
  
  if ((err = snd_pcm_hw_params_set_channels (Obj->handle, hw_params, 1)) < 0) {
    return(-1);
  }

  if ((err = snd_pcm_hw_params_set_rate_near (Obj->handle, hw_params, &rate, 0)) < 0) {
    return(-1);
  }

  /* set the buffer time */
  if((err = snd_pcm_hw_params_set_buffer_time_near(Obj->handle, 
						   hw_params, &buffer_time, &dir)) < 0 ) {
    return(-1);
  }

  if((err = snd_pcm_hw_params_get_buffer_size(hw_params, &Obj->buffer_size)) < 0 ) {
    return(-1);
  }
	
  if((err = snd_pcm_hw_params_set_period_time_near(Obj->handle, hw_params, 
						   &period_time, &dir)) < 0 ) {
    return(-1);
  }

  if((err = snd_pcm_hw_params_get_period_size(hw_params, &Obj->period_size, &dir)) < 0 ) {
    return(-1);
  }

  if ((err = snd_pcm_hw_params (Obj->handle, hw_params)) < 0) {
    return(-1);
  }

  snd_pcm_hw_params_free (hw_params);

  if ((err = snd_pcm_sw_params_malloc (&Obj->sw_params)) < 0) {
    return(-1);
  }

  if ((err = snd_pcm_sw_params_current (Obj->handle, Obj->sw_params)) < 0) {
    return(-1);
  }

  if ((err = snd_pcm_sw_params_set_avail_min (Obj->handle, 
					      Obj->sw_params,Obj->period_size )) < 0) {
    return(-1);
  }
	
  if ((err = snd_pcm_sw_params_set_start_threshold (Obj->handle,Obj->sw_params, 
						    Obj->buffer_size - Obj->period_size)) < 0) {
    return(-1);
  }

  if ((err = snd_pcm_sw_params (Obj->handle, Obj->sw_params)) < 0) {
    return(-1);
  }
  return(0);
}

/**********************************************************
  Store PCM Data to Sigle Buffer
**********************************************************/
static void StorePCMDataToBuffer(PAlsaObj Obj)
{
  if( Obj->Read != Obj->Write ) {
    for( ; Obj->Size < Obj->MaxSize ;  ) {
      *(Obj->buf + Obj->Size++) = *(Obj->Read->Buf + Obj->Read->Written++);

      if( Obj->Read->Written >= Obj->Read->Size ) {
 
	Obj->Read->Written = Obj->Read->Size = 0;
	Obj->Read = Obj->Read->Next;
	Obj->Used++;
	if( Obj->Read == Obj->Write ) {
	  break;
	}
      }
    }
  }
}

/********************************************************
  Store Last Sound Data
********************************************************/
static int LastSoundData = 0;

/********************************************************
  Write PCM Data to ALSA Raw
********************************************************/
static int WritePCMRow(PAlsaObj Obj,int aaa)
{
  int err,state;
  short int *LL;

  state = snd_pcm_state(Obj->handle);
  err = 0;
  if( state != 0x04 ) {
    if( (err = snd_pcm_writei(Obj->handle,
			      (short int *)Obj->buf,
			      Obj->period_size)) == Obj->period_size ) {


      LL = (short int *)Obj->buf + aaa - 2;
      LastSoundData = *LL;
      
      Obj->Size -= aaa;

      if( Obj->Size > 0 ) {
	/* $BM>$C$F$$$kItJ,$r%3%T!<$9$k(B */
	memcpy((char *)Obj->buf,(char *)Obj->buf + aaa,Obj->Size);

	/* Add New */
	StorePCMDataToBuffer(Obj);
      }
    }
  }
  return(err);
}

/**********************************************************
  Write NULL Sound
**********************************************************/
static void WriteNullSound(PAlsaObj Obj,int aaa)
{
  short  Buf[AUDIO_44100_100_NUM];
  int i;
  short int DD;
  double Sa,LastD;

  //fprintf(stderr,"Null Buf Size: %d\n",aaa);

  if( LastSoundData == 0 ) {
    DD = 0;
    for( i = 0 ; i < aaa ; i++ ) {
      Buf[i] = 0;
    }
    snd_pcm_writei(Obj->handle,(short int *)Buf,aaa);
    return;
  }

  LastD = (double)LastSoundData;
  Sa = (double)abs(LastSoundData) / (double)aaa;

  //fprintf(stderr,"Last %d Sa:%f\n",LastSoundData,Sa);

  if( LastSoundData > 0 ) {
    LastD = -LastD + Sa;
  }
  else {
    LastD = -LastD - Sa;
  }
  
  for( i = 0 ; i < aaa ; i++ ) {
    DD = (short)LastD;
    Buf[i] = DD;
    //snd_pcm_writei(Obj->handle,(short int *)&DD,2);
    if( LastD > 0.0 ) {
      LastD = -LastD + Sa;
    }
    else if (LastD < 0.0 ) {
      LastD = -LastD - Sa;
    }
  }
  snd_pcm_writei(Obj->handle,Buf,aaa);
  LastSoundData = 0;
}


/**********************************************************
  Write PCM Data to ALSA Audio
**********************************************************/
static void WritePCMDataToAlsa(PAlsaObj Obj)
{
  snd_pcm_sframes_t aaa,avail;
  int               err;

  aaa   = Obj->period_size * 2;

  if( ! Obj->handle ) {
    return;
  }

  avail = snd_pcm_avail_update(Obj->handle);

  if( avail < Obj->period_size * 2 ) {
    return;
  }

  //fprintf(stderr,"avail:%d aaa:%d\n",avail,aaa);

  /* $B$3$N;~(BObj->Size $B$O!"(Bavail $B$h$jBg$-$/!"(B aaa $B$h$jBg$-$$(B */
  while(  Obj->Size >= avail ) {
    StorePCMDataToBuffer(Obj);

    //fprintf(stderr,"avail:%d \n",avail);

    if( Obj->Size >= aaa ) {
      if( (err = WritePCMRow(Obj,aaa)) >= 0 ) {
        avail = snd_pcm_avail_update(Obj->handle);
      }
      else {
        if( err == -EPIPE ) {
          snd_pcm_prepare (Obj->handle);

          if( (err = WritePCMRow(Obj,aaa)) >= 0 ) {
	    avail = snd_pcm_avail_update(Obj->handle);
          }
	  else {
	    return;
	  }
        }
	else if( err == -EAGAIN ) {
	  
	  if( (err = WritePCMRow(Obj,aaa)) >= 0 ) {
	    avail = snd_pcm_avail_update(Obj->handle);
          }
	  else {
	    return;
	  }
	}
	else {
	  return;
	}
      }
    }
    else {
      /* No Data in Buffer */
      WriteNullSound(Obj,avail);
      return;
    }
  }
  if( Obj->Size < avail ) {
    /* No Data in Buffer */
    WriteNullSound(Obj,avail);
  }
}

/*****************************************************
  Sound Object Check
*****************************************************/
static void SoundObjectCheck(PAlsaObj Obj)
{
  int      state;

  if( ! Obj ) {
    return;
  }
  
  if( ! Obj->handle ) {
    SetALSADrive(Obj);
  }

  state = snd_pcm_state(Obj->handle);

  if( state == 0x04 ) { 			/* Need Stop */
    if( Obj->handle ) {
      snd_pcm_sw_params_free(Obj->sw_params);
      snd_pcm_close(Obj->handle);
      SetALSADrive(Obj);
    }
    Obj->Start   = 0;
    Obj->StopIDR = Obj->StopIDW;
  }
}

/************************************************************
  Interface Functions
************************************************************/

/************************************************************
  Initialize ALSA Driver And Allocate Memory
  Call by GTK Main Thread
************************************************************/
static void * InitAlsaDriver(char *Name,char *Exiting,char *JoinRoom)
{
  PAlsaObj             Obj;
  PAudioBuffer         Top,Prv,Now;
  int                  i;

  if((Obj = (PAlsaObj)malloc(sizeof(AlsaObj))) == NULL ) {
    return((void *)NULL);
  }

  /* Copy Device Name */
  strcpy(Obj->DeviceName,Name);
  Obj->Exiting = Exiting;
  Obj->JoinRoom = JoinRoom;

  if( SetALSADrive(Obj) < 0 ) {
    return((void *)NULL);
  }
 
  Obj->MaxSize = Obj->period_size * 5 * snd_pcm_format_width(SND_PCM_FORMAT_S16_LE) / 8;
  Obj->Size    = 0;

  /* 500 Mill Sec Unit */
  if((Obj->buf = (char *)malloc(Obj->MaxSize)) == NULL ) {
    return((void *)NULL);
  }

  Obj->Start   = 0;
  Obj->Used    = Obj->Store  = 0;
  Obj->StopIDW = Obj->StopIDR = 0;

  Top = Prv = NULL;

  for( i = 0 ; i < MAX_AUDIO_NUM ; i++ ) {
    if((Now = (PAudioBuffer)malloc(sizeof(AudioBuffer))) == NULL ) {
      return((void *)NULL);
    }

    if((Now->Buf = (unsigned char *)malloc(AUDIO_44100_100_NUM)) == NULL ) {
      return((void *)NULL);
    }
    Now->Size = Now->Written = 0;

    if( ! Top ) {
      Top = Now;
    }

    if( Prv ) {
      Prv->Next = Now;
    }

    Prv = Now;
  }
  Now->Next = Top;

  /* Store Ring Buffer */
  Obj->Read = Obj->Write = Now;
  return((void *)Obj);
}

/***********************************************************
   Get Wirte Buffer
   Called by Network Thread
***********************************************************/
static unsigned char *GetWriteALSABuffer(void *OO)
{
  PAlsaObj            Obj;
  Obj = (PAlsaObj)OO;

  return((unsigned char *)Obj->Write->Buf);
}

/**********************************************************
  Set Next Write Buffer
  Called by Network Thread
**********************************************************/
static void SetNextALSAWriteBuffer(void *OO,int Size)
{
  PAlsaObj            Obj;
  Obj = (PAlsaObj)OO;
  if( ! Obj ) {
    return;
  }
  Obj->Write->Size = Size;
  Obj->Write->Written = 0;
  Obj->Write = Obj->Write->Next;
  Obj->Store++;
}

/**********************************************************
  Sound Function
  Called by sound Thread by g_thread_create (Each 10 mill sec 
  the thread calls this function)
**********************************************************/
static void SoundAlSAFunction(void *OO)
{
  PAlsaObj            Obj;
  Obj = (PAlsaObj)OO;

  if( ! Obj ) {
    return;
  }

  if( Obj->StopIDW != Obj->StopIDR ) {

    /* $BDd;_MW5a$,Mh$?(B */
    if( Obj->handle ) {
      snd_pcm_sw_params_free(Obj->sw_params);
      snd_pcm_close(Obj->handle);
      Obj->handle = NULL;
    }
		
    Obj->Read = Obj->Write;
    Obj->Read->Size = Obj->Read->Written = 0;
    Obj->Start   = 0;
    Obj->Size    = 0;
    Obj->StopIDR = Obj->StopIDW;
    Obj->Used    = Obj->Store = 0;
  }

  if( *Obj->Exiting == 0 && *Obj->JoinRoom ) {
    StorePCMDataToBuffer(Obj);

    /* ALSA $B$,@5$7$$$+$I$&$,D4$Y$k(B */
    SoundObjectCheck(Obj);

    if( Obj->Start == 0 ) {	

      if( (Obj->Used + 5) < Obj->Store  ) { 
	if( Obj->handle ) {
	  WritePCMDataToAlsa(Obj);
	  snd_pcm_start(Obj->handle);
	  Obj->Start = 1;
	}
      }
    }
    else {
      if( Obj->handle ) {
	WritePCMDataToAlsa(Obj);
      }
    }
  }
}

/********************************************************
  Stop ALSA Audio
  Called by GTK Main Thread
********************************************************/
static void StopALSAAudio(void *OO)
{
  PAlsaObj            Obj;
  Obj = (PAlsaObj)OO;
  Obj->StopIDW++;
}

/********************************************************
  Finish ALSA Audio
  Called by GTK Main Thread
********************************************************/
static void FinishALSAAudio(void *OO)
{
  PAlsaObj            Obj;
  int i;
  PAudioBuffer        Now,Next;

  Obj = (PAlsaObj)OO;
  if( Obj->handle ) {
    snd_pcm_sw_params_free(Obj->sw_params);
    snd_pcm_close(Obj->handle);
    Obj->handle = NULL;
  }
  Now = Obj->Read;
  for( i = 0 ; i < MAX_AUDIO_NUM ; i++ ) {
    Next = Now->Next;
    free((void *)Now);
    Now = Next;
  }
  free((void *)Obj);
}

/***************************************************************
  Set ALSA Driver

 This function called by dlsym(Handle,"SetALSADriver");
 So, propgramers can not change function mame and arguments.
 Called by GTK Main Thread.
***************************************************************/
int SetALSADriver(PSoundObj Obj,char *Name,char *Exiting,char *JoinRoom)
{
  if((Obj->Handle = InitAlsaDriver(Name,Exiting,JoinRoom)) == NULL ) {
    return(1);
  }
  Obj->GetAudioBuffer     = GetWriteALSABuffer;
  Obj->SetNextWriteBuffer = SetNextALSAWriteBuffer;
  Obj->SoundFunction      = SoundAlSAFunction;
  Obj->StopAuidoFunction  = StopALSAAudio;
  Obj->FinishAuidoFunction = FinishALSAAudio;
  return(0);
}

